import matplotlib.pyplot as plt
import numpy as np
from textwrap import wrap
from pandas import DataFrame


class PrepareData:
    @staticmethod
    def cleaning_data(all_alignment_reports, sequencing_report):
        all_alignment_reports = all_alignment_reports.sort_values(
            "Input file(s)"
        ).reset_index(drop=True)
        unique_experiments = sequencing_report.sort_values("Experiment")[
            "Experiment"
        ].unique()
        unique_experiments = list(unique_experiments)
        tot_sequencing_reads = all_alignment_reports["Total sequencing reads"]
        aligned_reads = (
            all_alignment_reports["Successfully aligned reads"]
            .str.split("(", expand=True)
            .iloc[:, 0]
        )
        data = DataFrame([unique_experiments, aligned_reads, tot_sequencing_reads]).T
        data.columns = ["Experiment", "Aligned_Reads", "tot_sequenced_reads"]
        return data


class AlignmentPlot:
    def __init__(
        self,
        sequencing_report,
        all_alignment_reports,
        ax=None,
        font_settings={},
        legend_settings={},
        apply_log=False,
    ) -> None:
        self.ax = ax
        self.data = PrepareData().cleaning_data(
            all_alignment_reports, sequencing_report
        )
        self.base_plot()
        self.log_x_axis(apply_log)
        self.add_labels(font_settings)
        self.add_legend()

    def base_plot(self):
        self.ax.bar(
            self.data.Experiment,
            np.array(self.data.tot_sequenced_reads).astype(np.float32),
            label="Total Sequenced Reads",
            color="lightsalmon",
            alpha=1,
        )
        self.ax.bar(
            self.data.Experiment,
            np.array(self.data.Aligned_Reads).astype(np.float32),
            label="Aligned Reads",
            color="lightskyblue",
            alpha=1,
        )
        plt.xticks(rotation=45, ha="right", size=12)

    @staticmethod
    def log_x_axis(apply_log):
        if apply_log:
            plt.yscale("log")

    def add_labels(self, font_settings):
        if font_settings != {}:
            plt.ylabel("Reads Count", **font_settings)
            plt.xlabel("Sample", **font_settings)
            original_fontsize = font_settings["fontsize"]
            font_settings["fontsize"] = 22
            title = "\n".join(wrap("Alignment Quality of the analyzed samples", 40))
            self.ax.set_title(title, pad=12, **font_settings)
            font_settings["fontsize"] = original_fontsize

    def add_legend(self, legend_settings):
        if legend_settings != {}:
            self.ax.legend(**legend_settings)


def barplot(
    ax,
    all_alignment_reports,
    sequencing_report_all,
    font_settings,
    legend_settings,
    apply_log,
):
    boxplot_data_frame = cleaning_data(all_alignment_reports, sequencing_report_all)
    ax.bar(
        boxplot_data_frame.Experiment,
        np.array(boxplot_data_frame.tot_sequenced_reads).astype(np.float32),
        label="Total Sequenced Reads",
        color="lightsalmon",
        alpha=1,
    )
    ax.bar(
        boxplot_data_frame.Experiment,
        np.array(boxplot_data_frame.Aligned_Reads).astype(np.float32),
        label="Aligned Reads",
        color="lightskyblue",
        alpha=1,
    )
    plt.xticks(rotation=45, ha="right", size=12)

    ax.legend(**legend_settings)
    plt.ylabel("Reads Count", **font_settings)
    plt.xlabel("Sample", **font_settings)

    if apply_log == True:
        plt.yscale("log")
    plt.tight_layout()
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22
    title = "\n".join(wrap("Alignment Quality of the analyzed samples", 40))
    ax.set_title(title, pad=12, **font_settings)
    font_settings["fontsize"] = original_fontsize


# matplotlib.use("Qt5Agg")
# boxplot_data_frame = cleaning_data(all_alignment_reports, sequencing_report_all)
# ax.bar(boxplot_data_frame.Experiment, np.array(boxplot_data_frame.tot_sequenced_reads).astype(np.float32),label="Total Sequenced Reads",color="orange", np.array(boxplot_data_frame.Aligned_Reads).astype(np.float32),
# label="Aligned Reads",
# color="royalblue")
# ax.set_xticks(rotation=45, ha = 'right', size=5)
# params_legend = openParams("USQ_plot_legend_params.txt")
# ax.legend(**params_legend)
# plot_style = openParams('plot_style.txt')
# ax.set_ylabel("Reads Count", **plot_style)
# ax.set_xlabel("Sample", **plot_style)
# if apply_log == True:
#   ax.set_yscale("log")
# fig.tight_layout()



from .set_env import VE
import subprocess
import os

class DeepAB:
    def __init__(self):
        self.git_available()
        Ve_Control = VE(name_ve = "deepab")
        self.clone_repo()
        Ve_Control.switch_to_ve()
        self.install_requirements()
        
        
    def clone_repo():
        if not os.path.isdir("DeepAb"):
            while user_input not in ["X", "x", ""]:
                user_input = input("You will now clone the repository from DeepAb.\nPlease make sure that you have read the corresponding licencing on their repository.\nPress enter to continue and press x to abort the process.")
                if user_input not in ["X", "x", ""]:
                    print("Please enter a correct value.")
            
            if user_input == "": 
                subprocess.run(["git",
                                "clone", 
                                "https://github.com/RosettaCommons/DeepAb.git"])
            else:
                pass
            
    def install_requirements():
        if os.path.isfile(os.path.join(os.getcwd(), "DeepAb", "requirements.txt")):
            subprocess.run(["pip",
                            "install", 
                            "-r", 
                            "DeepAb/requirements.txt"])
            
    




from src.ExpoSeq.pipeline import PlotManager
import pandas as pd
import random 

def test_pipeline():
    
    plot = PlotManager(experiment = "multi_region", module_dir=r"C:\Users\nilsh\my_projects\ExpoSeq\src\ExpoSeq\software_tests\test_files", allow_binding_data=False, test_version=True, show_df = False)
    assert "targetSequences" in plot.avail_regions
    plot.change_region(region = "aaSeqCDR2")
    plot.change_region(region = "targetSequences")
    plot.change_filter(length_threshold_aa=2, min_read_count=2, remove_gaps = False, remove_errors = True)
    seq_report = plot.sequencing_report
    assert "aaSeqtargetSequences" in seq_report.columns.to_list()
    assert "nSeqtargetSequences" in seq_report.columns.to_list()
    
    plot.print_antigens()
    plot.print_samples()
    assert plot.region_of_interest == "aaSeqtargetSequences"
    plot.discard_samples(["Pool1_R1_001"])
    seq_report = plot.sequencing_report
    assert plot.sequencing_report.shape[0] == 0
    
    plot = PlotManager(experiment = "test_show", module_dir = "src/ExpoSeq/software_tests/test_files", allow_binding_data=False, test_version=True, show_df = False)

    plot.lengthDistribution_single()
    plot.lengthDistribution_multi()
    plot.print_antigens()
    plot.print_samples()
    plot.aa_distribution()
    plot.rarefraction_curves()
    plot.logoPlot()
    plot.logoPlot_multi()
    plot.rel_seq_abundance()
    plot.basic_cluster(batch_size = 100)
    plot.embedding_tsne(batch_size = 100)
    plot.morosita_horn()
    plot.jaccard()
    plot.sorensen()
    plot.relative()
    plot.levenshtein_dendrogram()
    plot.alignment_quality()
    plot.umap_sample_cluster(batch_size = 51)
    plot.umap_clustering_characteristic(batch_size = 51)
    
    plot = PlotManager(experiment = "test_show", module_dir = "src/ExpoSeq/software_tests/test_files", test_version=True, allow_binding_data=False, show_df = False)
    sequencing_report = plot.sequencing_report
    sample_0 = sequencing_report[sequencing_report["Experiment"] == "GeneMind_1"]
    aa = sample_0["aaSeqCDR3"]
    top_ten = aa.to_list()[:10]
    random_list = [random.randint(1, 3345832) for _ in range(10)]
    binding_data = pd.DataFrame(data = {"aaSeqCDR3": top_ten, "Antigen 0": random_list})
    plot.binding_data = binding_data
    plot.preferred_antigen= "Antigen 0"
    plot.dendro_bind(antigens = ["Antigen 0"])
    
   # plot.tsne_cluster_AG()
    plot.sample_diversity()
    plot.cluster_one_AG()
    plot.connect_samples()
    plot.embedding_network(batch_size = 100)
    
    plot = PlotManager(experiment = "test_show", module_dir = "src/ExpoSeq/software_tests/test_files", allow_binding_data="src/ExpoSeq/software_tests/test_files/binding_data.csv", test_version=True, show_df = False)
    plot.cluster_binding_data(batch_size = 100, iterations_tsne = 251)
    plot.ls_distance_binding()
    plot.tsne_cluster_AG(iterations_tsne = 251)
    sample_names = plot.experiments_list
    plot.lengthDistribution_single(sample = sample_names[0])
    samples_multi = [sample_names[0], sample_names[1]]
    plot.lengthDistribution_multi(samples = samples_multi)
    plot.aa_distribution(sample = [sample_names[0]], region = [3,7])
    plot.aa_distribution(sample = [sample_names[0]], region = [3,7], protein = False)
    plot.rarefraction_curves(samples = samples_multi)
    plot.logoPlot(sample = [sample_names[0]], highlight_specific_pos=4, chosen_seq_length = 10)
    plot.logoPlot(color_scheme = "hydrophobicity")
    plot.logoPlot(color_scheme = "charge")
    plot.logoPlot(color_scheme = "chemistry")
    plot.logoPlot(color_scheme = "NajafabadiEtAl2017")
    plot.logoPlot(color_scheme="dmslogo_charge")
    plot.logoPlot(color_scheme = "skylign_protein")
    plot.sorensen(specific_experiments = samples_multi)
    plot.rel_seq_abundance(sample = sample_names[0], alpha_val = 0.5, top_clone_fraction = 0.5)
    plot.cluster_binding_data_umap(batch_size = 51)
       

    


import networkx as nx
import math
from textwrap import wrap
import numpy as np
from .tidy_protbert_embedding import TransformerBased
import matplotlib.pyplot as plt
import community.community_louvain as community
from sklearn.cluster import KMeans


class PrepareData:
    @staticmethod
    def input_check(model, list_experiments, batch_size):
        model_types_all = [
            "Rostlab/ProstT5_fp16",
            "Rostlab/prot_t5_xl_uniref50",
            "Rostlab/prot_t5_base_mt_uniref50",
            "Rostlab/prot_bert_bfd_membrane",
            "Rostlab/prot_t5_xxl_uniref50",
            "Rostlab/ProstT5",
            "Rostlab/prot_t5_xl_half_uniref50-enc",
            "Rostlab/prot_bert_bfd_ss3",
            "Rostlab/prot_bert_bfd_localization",
            "Rostlab/prot_electra_generator_bfd",
            "Rostlab/prot_t5_xl_bfd",
            "Rostlab/prot_bert",
            "Rostlab/prot_xlnet",
            "Rostlab/prot_bert_bfd",
            "Rostlab/prot_t5_xxl_bfd",
        ]
        assert (
            model in model_types_all
        ), f"Please enter a valid model name which are\n{model_types_all}. You can find nearly all of the models at: https://huggingface.co/Rostlab"
        assert (
            type(list_experiments) == list
        ), "You have to give a list with the samples you want to analyze"
        assert (
            type(batch_size) == int
        ), "You have to give an integer as input for the batch_size"

    def cleaning(
        self,
        sequencing_report,
        model_choice,
        batch_size,
        list_experiments,
        region_of_interest,
    ):
        self.input_check(model_choice, list_experiments, batch_size)
        Transformer = TransformerBased(choice=model_choice)
        sequences, sequences_filtered, selected_rows = Transformer.filter_sequences(
            sequencing_report,
            batch_size=batch_size,
            experiments=list_experiments,
            region_of_interest=region_of_interest,
            binding_data=None,
        )
        sequences_filtered = sequences_filtered.to_list()
        sequences_list = Transformer.embedding_per_seq(sequences, normalize=True)
        return sequences_list, sequences_filtered, selected_rows


class Network_Embedding:
    def __init__(
        self,
        ax,
        sequencing_report,
        list_experiments,
        model_choice,
        batch_size,
        font_settings,
        cmap=plt.cm.RdYlBu,
        nodesize=None,
        threshold_distance=1,
        region_of_interest="aaSeqCDR3",
    ) -> None:
        self.ax = ax
        sequences_list, sequences_filtered, selected_rows = PrepareData().cleaning(
            sequencing_report,
            model_choice,
            batch_size,
            list_experiments,
            region_of_interest,
        )
        self.G = nx.Graph()
        self.add_edges(sequences_list, sequences_filtered, threshold_distance)
        if nodesize == None:
            nodesize = self.add_nodesize(selected_rows, region_of_interest)
        else:
            nodesize = 500
        self.generate_plot(nodesize=nodesize, cmap=cmap)
        title = "\n".join(wrap("Network embedding for given samples", 40))
        self.ax.set_title(title, pad=12, **font_settings)

    @staticmethod
    def k_means_clustering(distance_matrix):
        """
        This function follows the elbow methods to determine the number of clusters
        Args:
            distance_matrix (_type_): _description_
        """
        sse = []  # sum of squared errors
        sequence_size = distance_matrix.shape[0]
        max_cluster_number = sequence_size // 3
        list_k = list(range(1, max_cluster_number))

        for k in list_k:
            km = KMeans(n_clusters=k)
            km.fit(distance_matrix)
            sse.append(km.inertia_)
        differences = np.diff(sse)

    def add_edges(self, sequences_list, sequences_filtered, threshold_distance):
        num_sequences = sequences_list.shape[0]
        distance_matrix = np.zeros((num_sequences, num_sequences))
        for seq_1 in range(num_sequences):
            for seq_2 in range(seq_1 + 1, num_sequences):
                p = sequences_list[seq_1]
                q = sequences_list[seq_2]
                distance = np.sqrt(
                    np.sum(abs(sequences_list[seq_1] - sequences_list[seq_2]) ** 2)
                )  # euclidian distance of embedding coordinates for the 1024 dimensions
                if distance < threshold_distance / num_sequences:
                    string1 = sequences_filtered[seq_1]
                    string2 = sequences_filtered[seq_2]
                    self.G.add_edge(string1, string2, weight=distance)

    def add_nodesize(self, selected_rows, region_of_interest):
        nodesize = []
        label_numbers = {}
        label_sequences = {}
        for index, g in enumerate(self.G):

            for i in selected_rows[region_of_interest]:

                if i == g:
                    node = math.sqrt(
                        selected_rows.loc[
                            selected_rows[region_of_interest] == i, "cloneFraction"
                        ].values[0]
                    )
                    nodesize.append(node)
                    if node * 500 > 1 / 100 * 500:
                        label_numbers[g] = index
                        label_sequences[g] = g
                    else:
                        label_numbers[g] = ""
                        label_sequences[g] = ""
                    break
        for y, n in enumerate(nodesize):
            nodesize[y] = int(n * 500)
        return nodesize

    def generate_plot(self, nodesize, cmap):
        partition = community.best_partition(self.G)
        pos = nx.spring_layout(self.G)
        nx.draw_networkx_nodes(
            self.G, pos, cmap=cmap, node_color=list(partition.values())
        )
        nx.draw_networkx_edges(self.G, pos, alpha=1)



import editdistance
import networkx as nx
import math
import seaborn as sns
import community
import matplotlib.pyplot as plt


class PrepareData:
    def __init__(self, summed_clonefraction, max_num_reads):
        self.summed_clonefraction = summed_clonefraction
        self.max_num_reads = max_num_reads

    def get_top_10_percent(self, group):
        sorted_group = group.sort_values(by="cloneFraction", ascending=False)
        cumulative_sum = sorted_group["cloneFraction"].cumsum()
        return sorted_group[cumulative_sum <= self.summed_clonefraction][
            : self.max_num_reads
        ]

    def cleaning(self, report):
        grouped = report.groupby("Experiment")
        result_df = grouped.apply(self.get_top_10_percent)
        # Step 4: Reset index (if needed)
        result_df = result_df.reset_index(drop=True)
        report = result_df
        return report


class ClusterExperiment:
    def __init__(
        self,
        sequencing_report,
        region_string,
        summed_clonefraction,
        max_num_reads,
        ax=None,
        edge_color="peachpuff",
        max_weight_lines=100,
    ):
        report = sequencing_report
        self.region_string = region_string
        report = self.prepare_data(
            sequencing_report, summed_clonefraction, max_num_reads
        )
        self.G = nx.Graph()
        distances = self.nodes_and_edges(report, max_weight_lines)
        nodesize = self.customizations(report)
        sample_colors, node_colors = self.create_colors(report)
        if ax != None:
            self.draw_plot(nodesize, node_colors, ax, edge_color)
            self.legend(report, sample_colors)

    @staticmethod
    def prepare_data(sequencing_report, summed_clonefraction, max_num_reads):
        report = PrepareData(summed_clonefraction, max_num_reads).cleaning(
            sequencing_report
        )
        return report

    def nodes_and_edges(self, report, max_weight_lines=100):
        distances = []
        for sample, group in report.groupby("Experiment"):
            sequences = group[self.region_string].tolist()
            for i in range(len(sequences)):
                seq1 = sequences[i]
                # Add node for seq1
                self.G.add_node(seq1, sample=sample)

                for j in range(i + 1, len(sequences)):
                    seq2 = sequences[j]
                    distance = editdistance.distance(seq1, seq2)

                    self.G.add_node(seq2, sample=sample)
                    distances.append((sample, seq1, seq2, distance))
                    # Add nodes for seq2 and connect them with an edge weighted by the Levenshtein distance
                    if distance != 0:
                        if distance > 5:
                            self.G.add_edge(
                                seq1,
                                seq2,
                                weight=int(max_weight_lines * 1 / (distance**2)),
                            )

                    else:
                        self.G.add_edge(seq1, seq2, weight=max_weight_lines)
        return distances

    def customizations(self, report):
        nodesize = []
        label_numbers = {}
        label_sequences = {}

        for index, g in enumerate(self.G):
            for i in report[self.region_string]:
                if i == g:
                    node = math.sqrt(
                        report.loc[
                            report[self.region_string] == i, "cloneFraction"
                        ].values[0]
                    )
                    nodesize.append(node)
                    if node * 500 > 1 / 100 * 500:
                        label_numbers[g] = index
                        label_sequences[g] = g
                    else:
                        label_numbers[g] = ""
                        label_sequences[g] = ""
                    break
        for y, n in enumerate(nodesize):
            nodesize[y] = int(n * 500)
        max_fraction_node = max(label_numbers, key=label_numbers.get)
        #   label_numbers = {node: index if node == max_fraction_node else "" for node, index in label_numbers.items()}
        #  label_sequences = {node: seq if node == max_fraction_node else "" for node, seq in label_sequences.items()}

        return nodesize

    def create_colors(self, report):
        num_samples = len(report["Experiment"].unique())
        # Generate a color palette with num_samples distinct colors
        palette = sns.color_palette("hsv", num_samples)
        # Create a dictionary to map samples to colors
        sample_colors = {
            sample: color
            for sample, color in zip(report["Experiment"].unique(), palette)
        }
        node_colors = [sample_colors[self.G.nodes[n]["sample"]] for n in self.G.nodes]
        return sample_colors, node_colors

    def draw_plot(self, nodesize, node_colors, ax, edge_color):
        G = self.G
        partition = community.best_partition(G)

        weight = 0.1

        pos = nx.spring_layout(G)
        nx.draw_networkx_nodes(
            self.G, pos, node_size=nodesize, node_color=node_colors, ax=ax
        )
        nx.draw_networkx_edges(self.G, pos, alpha=0.1, edge_color=edge_color, ax=ax)

    def legend(self, report, sample_colors):
        unique_samples = report["Experiment"].unique()

        # Create a list of colors corresponding to the samples
        colors = [sample_colors[sample] for sample in unique_samples]

        # Create a legend mapping sample names to colors
        legend = {sample: color for sample, color in zip(unique_samples, colors)}

        # Add a legend to the plot
        plt.legend(
            handles=[
                plt.Line2D(
                    [0],
                    [0],
                    marker="o",
                    color="w",
                    label=sample,
                    markersize=10,
                    markerfacecolor=color,
                )
                for sample, color in legend.items()
            ],
            loc="center left",
            bbox_to_anchor=(1, 0.5),
        )



import os
import subprocess
import pandas as pd
from ExpoSeq.settings.full_sequence_finder import FullSequence
from difflib import SequenceMatcher
import numpy as np

class ExtractFasta:
    def __init__(self, merged_fasta_file: str, save_dir):
        assert merged_fasta_file.endswith(".fasta") or merged_fasta_file.endswith(".fa")
        self.merge_fasta_file = merged_fasta_file
        self.save_dir = self.check_create_dir(save_dir)

    @staticmethod
    def check_create_dir(dir):
        save_dir = os.path.join(dir, "binder_import")
        if not os.path.exists(save_dir):
            os.makedirs(os.path.join(dir, "binder_import"))
        else:
            pass

        return save_dir

    def merge_fasta_files(self, dir):
        with open(self.merge_fasta_file, "w") as oh:
            first_line = True
            for f in os.listdir(dir):
                if f.endswith(".fasta") or f.endswith(".fa") or f.endswith(".fas"):
                    with open(os.path.join(dir, f)) as fh:
                        lines = fh.readlines()
                        for i, line in enumerate(lines):
                            if line.startswith(">"):
                                if first_line:
                                    oh.write(line)
                                    first_line = False
                                else:
                                    oh.write("\n" + line)
                            else:

                                oh.write(line)


    def align_mixcr(self, mixcr_path: str, preset_name="exom-full-length", species=""):
        """_summary_

        Args:
            mixcr_path (str): _description_
            preset_name (str, optional): _description_. Defaults to "milab-human-bcr-multiplex-full-length".
            species (str, optional): For available species check out:https://github.com/repseqio/library-imgt/releases . Defaults to "".
        """
        command = [
            "java",
            "-jar",
            mixcr_path,
            "analyze",
            preset_name,
            self.merge_fasta_file,
            os.path.join(self.save_dir, "result_mixcr_fasta"),
            "-f",
            "--no-json-reports",

        ]
        if species == "":
            pass
        else:
            command.append("--species")
            command.append(f"{species}")
        subprocess.run(command)

    @staticmethod
    def get_headers_and_sequences_from_fasta(fasta_file):
        headers = []
        sequences = []
        sequence = ""
        with open(fasta_file, "r") as file:
            for line in file:
                if line.startswith(">"):
                    headers.append(line.strip())
                    if sequence != "":
                        sequences.append(sequence)
                        sequence = ""
                else:
                    sequence += line.strip()
            sequences.append(sequence)  # for the last sequence
            
        df = pd.DataFrame({"Header": headers, "targetSequences": sequences})
        return df
    @staticmethod
    def sequence_in_other(row, df_long, supply_possible_cols = None):
        """_summary_

        Args:
            row (_type_): row from iteration from tsv from mixcr output.
            df_long (_type_): Sequences and headers from original fasta file

        Returns:
            _type_: _description_
        """
        if supply_possible_cols is not None:
            possible_cols = supply_possible_cols
        # Find the row in df_long where the 'nSeqtargetSequences' value is the same as in the current row
        for i, seq in enumerate(df_long["targetSequences"]):
            if row["nSeqCDR3"] in seq:
                matching_row = df_long.iloc[i, :]
                possible_header = matching_row.loc['Header']
                if supply_possible_cols is not None:
                    a = possible_header.split("_")
                    b = "_".join(a[:3])

                    modified_header = b[1:] + "_M"
                    if modified_header in possible_cols:
                        return modified_header
                    else:
                        pass
        # If no matching row was found, return NaN
        return np.nan


    @staticmethod   
    def translate_sequence(amino_acid_sequence):
        # Define the codon table
        codon_table = {
            "F": "TTT",
            "L": "TTA",
            "I": "ATT",
            "M": "ATG",
            "V": "GTT",
            "S": "TCT",
            "P": "CCT",
            "T": "ACT",
            "A": "GCT",
            "Y": "TAT",
            "H": "CAT",
            "Q": "CAA",
            "N": "AAT",
            "K": "AAA",
            "D": "GAT",
            "E": "GAA",
            "C": "TGT",
            "W": "TGG",
            "R": "CGT",
            "G": "GGT",
            "*": "TAA",
            "_": "",
            
        }

        # Translate the amino acid sequence to a nucleotide sequence
        nucleotide_sequence = "".join([codon_table[aa] for aa in amino_acid_sequence])

        return nucleotide_sequence

        
    def merge_and_read_seq(
        self,
        tsv_filename = None, 
        naming_col = None
    ):
        all_seqs = self.get_headers_and_sequences_from_fasta(self.merge_fasta_file)
        if tsv_filename is None:
            tsv_filename = os.path.join(
                self.save_dir, "result_mixcr_fasta" + ".clones_IGH" + ".tsv"
            )
        else:
            tsv_filename = tsv_filename
        tsv = pd.read_table(tsv_filename)

        #    tsv_merge = tsv_merge[~tsv_merge.apply(lambda row: row.astype(str).str.contains('region not covered').any(), axis=1)]
        aaSeq_columns = [col for col in tsv.columns if col.startswith("aaSeq")]
        aaSeq_columns = [
            col.replace("aaSeq", "") for col in tsv.columns if col.startswith("aaSeq")
        ]
        row_indexes = FullSequence(avail_regions=aaSeq_columns).find_connecting_seq()
        avail_regions = ["".join("aaSeq" + region) for region in row_indexes]
        tsv["merged_regions"] = tsv[avail_regions].apply(
            lambda row: "".join(row.values.astype(str)), axis=1
        )
        tsv = tsv[~tsv['merged_regions'].str.contains('region_not_covered')]
        tsv["nSeqtargetSequences"] = tsv["merged_regions"].apply(self.translate_sequence)
        tsv['Header'] = tsv.apply(lambda row: self.sequence_in_other(row, df_long=all_seqs, supply_possible_cols=naming_col), axis=1)
        tsv = tsv[["Header", "merged_regions"]]
        tsv = tsv[["Header", "merged_regions"]].dropna(subset=["Header"])
        tsv = tsv.rename(columns={"merged_regions": "aaSeqtargetSequences"})
        cols = list(tsv.columns)
        cols.insert(0, cols.pop(cols.index("aaSeqtargetSequences")))
        print(tsv)
        tsv = tsv.loc[:, cols]
        
        
        tsv.to_csv(tsv_filename.replace(".tsv", "_merged.csv"))




import pandas as pd
import numpy as np



class PrepareData:
    @staticmethod
    def find_seq_matches(sequencing_report,region_of_interest, protein, specific_experiments):
        if specific_experiments != False:
            sequencing_report = sequencing_report[sequencing_report['Experiment'].isin(specific_experiments)]
        else:
            pass
        unique_experiments = sequencing_report["Experiment"].unique()
        if protein == True:
            strand_column = region_of_interest
        else:
            strand_column =  region_of_interest
        unique_sequences = pd.DataFrame(sequencing_report[strand_column].unique())
        unique_sequences.rename(columns={0: strand_column},
                                inplace=True)

        for i, experiment in enumerate(unique_experiments):
            local_data = sequencing_report[sequencing_report["Experiment"] == experiment][strand_column].unique()
            local_data = pd.DataFrame(local_data, columns=[strand_column])
            unique_sequences = pd.merge(unique_sequences, local_data,
                                        on=strand_column,
                                        how='left',
                                        indicator=True)
            unique_sequences[experiment] = unique_sequences["_merge"].eq("both")
            unique_sequences = unique_sequences.drop("_merge", axis=1)
        unique_sequences.drop(strand_column,
                            inplace=True,
                            axis=1)
        unique_sequences = unique_sequences.astype("int")
        heatmap_axis = len(unique_experiments)
        return heatmap_axis, unique_sequences, unique_experiments

    def cleaning_jaccard(self, sequencing_report, region_of_interest, protein, specific_experiments):
        heatmap_axis, unique_sequences, unique_experiments = self.find_seq_matches(sequencing_report,
                                                                            region_of_interest,
                                                                            protein,
                                                                            specific_experiments)
        heatmap_absolute_jaccard = np.zeros([heatmap_axis, heatmap_axis])
        for index_sample_one in range(heatmap_axis):
            d1 = unique_sequences.iloc[:, index_sample_one]
            for index_sample_two in range(heatmap_axis):
                d2 = unique_sequences.iloc[:, index_sample_two]
                lib_matches = d1 * d2
                lib_matches = lib_matches.sum()
                a = lib_matches
                b = d1.sum() - lib_matches
                c = d2.sum() - lib_matches
                heatmap_absolute_jaccard[index_sample_one, index_sample_two] = lib_matches/(lib_matches+b+c)
                heatmap_absolute_jaccard[index_sample_two, index_sample_one] = lib_matches / (lib_matches + b + c)
        matrix = pd.DataFrame(heatmap_absolute_jaccard,
                            index = list(unique_experiments),
                            columns = list(unique_experiments))

        return matrix, unique_experiments




import subprocess
import os

class WindowsInstall:    
    def start_wsl():
        filepath = os.path.join(os.path.dirname(__file__), "installation_windows.ps1")
        subprocess.run(["powershell.exe", "-File", filepath])
    
    def exit_wsl():
        subprocess.run(["wsl --shutdown"])
        
    def wsl():
        subprocess.run(["wsl"])
        
    @staticmethod
    def create_install_dir():
        subprocess.run(["wsl", "mkdir", "install_lcf"])
        
    @staticmethod
    def install_lcf():
        subprocess.run(["wsl", "fsutil", "file", "SetCaseSensitiveInfo", ".", "enable"])
        subprocess.run(["wsl", "wget", "https://raw.githubusercontent.com/YoshitakaMo/localcolabfold/main/install_colabbatch_linux.sh"])
        subprocess.run(["bash", "install_colabbatch_linux.sh"])
        

WindowsInstall.start_wsl()

    


import glob
import os
from .markdown_builder import QuartoBuilder
import subprocess


class Subplotter:
    def __init__(self,  figure_title):
        self.check_base_dir()
        self.files = []
        self.figure_title = figure_title
        self.captures = []
        
    @staticmethod
    def check_base_dir():
        if os.path.isdir("tmp_quarto"):
            pass
        else:
            os.mkdir("tmp_quarto")
            
            
    def update_files(self, figure_dir):
        files_raw = glob.glob(os.path.join(figure_dir, "*png"))
        self.files = []
        for file in files_raw:
            self.files.append(os.path.abspath(file))
            
    def update_capture(self, fig_capture):
        self.captures.append(fig_capture)
        
    def add_as_subplot(self, fig, fig_capture = "", dir = "tmp_quarto" ):
        self.update_capture(fig_capture)
        if dir == "tmp_quarto":
            figure_dir = os.path.join(dir, self.figure_title)
        else:
            figure_dir = dir
            assert os.path.isdir(figure_dir)
        if os.path.isdir(figure_dir):
            pass
        else:
            os.makedirs(figure_dir)
        no = str(len(self.files))
        fig.savefig(os.path.join(figure_dir, f"{no}_.png"),
                                                     format = "png",
                                                     dpi = 300, 
                                                     bbox_inches='tight')
        self.update_files(figure_dir)
        
    def make_figure(self, dir = "tmp_quarto", ):
        Builder = QuartoBuilder(self.figure_title, figure=True)  
        Builder.add_subplot_figures(self.files, self.captures)
        
        if dir == "tmp_quarto":
            qmd_file = os.path.join(dir, self.figure_title, f"{self.figure_title}.qmd")
            Builder.write_quarto(save_dir=qmd_file)
        else:
            qmd_file = dir
            Builder.write_quarto(save_dir=qmd_file)
        subprocess.run(["quarto", "render", os.path.join(qmd_file, Builder.title + ".qmd")])







import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
from textwrap import wrap

class PrepareData:
    @staticmethod
    def prepare_unique_seq_table(sequencing_report, region_of_interest, protein = True, specific_experiments = False):
        if specific_experiments != False:
            sequencing_report = sequencing_report[sequencing_report['Experiment'].isin(specific_experiments)]
        else:
            pass

        if protein == True:

            strand_column =  region_of_interest
        else:
            strand_column = region_of_interest


        unique_experiments = sequencing_report["Experiment"].unique()

        unique_sequences = pd.DataFrame(sequencing_report[strand_column].unique()) #nseqcdr3 has to be changed since it is a chosen name for column

        unique_sequences.rename(columns={0: strand_column},
                                inplace=True)
        unique_experiments = list(unique_experiments)
        for experiment in unique_experiments:

            local_data = sequencing_report[sequencing_report["Experiment"] == experiment][[strand_column,
                                                                                        "cloneFraction"]]
            local_data = local_data.sort_values(by="cloneFraction", ascending=False)

            # Calculate the cumulative sum of clone fractions
            local_data["cumulativeFraction"] = local_data["cloneFraction"].cumsum()

            # Filter rows until the cumulative fraction reaches the threshold
            filtered_data = local_data[local_data["cumulativeFraction"] <= 0.95]

            local_data = filtered_data[[strand_column,"cloneFraction"]]

            unique_sequences = unique_sequences.merge(local_data,
                                                    how='left',
                                                    on = strand_column)
        # unique_sequences = unique_sequences.fillna(0)
            unique_sequences = unique_sequences.rename(columns={"cloneFraction": "cloneFraction" + experiment})
            # Convert column 'A' to float16
            unique_sequences[ "cloneFraction" + experiment] = unique_sequences[ "cloneFraction" + experiment].astype('float16')
        unique_sequences.drop(strand_column,
                            inplace=True,
                            axis=1)
        unique_sequences = unique_sequences.fillna(0)
        unique_sequences.reset_index(inplace = True)
        unique_sequences.drop('index', axis=1, inplace=True)
        return unique_sequences, unique_experiments
    
    @staticmethod
    def calc_matrix_values(unique_sequences, unique_experiments):
        columns = len(unique_experiments)
        morosita_horn_matrix = np.zeros((columns,
                                        columns))
        for index_sample_one in range(columns):
            d1 = unique_sequences.iloc[:, index_sample_one]
            start = 0
            for index_sample_two in range(columns):
                d2 = unique_sequences.iloc[:, index_sample_two]
                product_d1_d2 = np.array(d1) * np.array(d2)
                d1_square = np.array(d1)**2
                d2_square = np.array(d2)**2

                MH = (2*np.sum(product_d1_d2)) /(np.sum(d1_square) + np.sum(d2_square))
                morosita_horn_matrix[index_sample_one, index_sample_two] = MH
                morosita_horn_matrix[index_sample_two, index_sample_one] = MH
                start += 1
        matrix = pd.DataFrame(morosita_horn_matrix,
                            index = list(unique_experiments),
                            columns = list(unique_experiments))
        return matrix
        
    
    def cleaningPlot(self, samples, sequencing_report, protein, region_string, ):
        unique_sequences, unique_experiments = self.prepare_unique_seq_table(sequencing_report,
                                                                                region_string,
                                                                                protein = protein,
                                                                                specific_experiments = samples,
                                                                                )
        matrix = self.calc_matrix_values(unique_sequences,
                                        unique_experiments)
        unique_experiments = list(unique_experiments)
        matrix = matrix.sort_index(axis=1)
        matrix = matrix.sort_index(axis = 0)
        return matrix, unique_experiments
    





import numpy as np
import matplotlib.pyplot as plt
from ..settings.layout_finder import best_layout
from textwrap import wrap


class LengthDistributionSingle:
    def __init__(
        self,
        sequencing_report,
        sample,
        region_of_interest,
        ax=None,
        font_settings={},
        title_type="single",
    ):
        unique_length, counts_length = self.tidy(
            sequencing_report, sample, region_of_interest
        )
        self.plot(unique_length, counts_length, sample, ax, font_settings, title_type)

        if title_type == "single":
            self.title(sample, font_settings)
        else:
            pass

    @staticmethod
    def tidy(sequencing_report, sample, region_string):
        batch = sequencing_report[sequencing_report["Experiment"] == sample]
        length = batch[region_string].str.len()
        unique_length, counts_length = np.unique(np.array(length), return_counts=True)
        return unique_length, counts_length

    @staticmethod
    def plot(
        unique_length, counts_length, sample, ax, font_settings, title_type="single"
    ):
        ax.bar(
            unique_length, counts_length, color="lightskyblue"
        )  # Or whatever you want in the subplot
        # ax.set_xticks(range(0, max_length + 1, 1), range(0, max_length + 1, 1))
        ax.title.set_text(sample)
        if title_type == "single":
            ax.title.set_size(18)
        else:
            ax.title.set_size(12)
        ax.set_ylabel("Read Count", **font_settings)  # Y label
        ax.set_xlabel("Read Length", **font_settings)  # X label

        original_fontsize = font_settings["fontsize"]
        font_settings["fontsize"] = 20
        title = "\n".join(wrap("Length Distribution of " + sample, 40))
        plt.title(title, pad=12, **font_settings)
        font_settings["fontsize"] = original_fontsize

    @staticmethod
    def title(sample, font_settings):
        original_fontsize = font_settings["fontsize"]
        font_settings["fontsize"] = 20
        title = "\n".join(wrap("Length Distribution of " + sample, 40))
        plt.title(title, pad=12, **font_settings)
        font_settings["fontsize"] = original_fontsize


def length_distribution_multi(
    fig,
    sequencing_report,
    samples,
    font_settings,
    region_string,
    test_version=False,
):
    if samples == "all":
        unique_experiments = sequencing_report["Experiment"].unique()
        unique_experiments = np.sort(unique_experiments)
    else:
        sequencing_report = sequencing_report[
            sequencing_report["Experiment"].isin(samples)
        ]
        unique_experiments = sequencing_report["Experiment"].unique()
        unique_experiments = np.sort(unique_experiments)
    Tot = unique_experiments.shape[0]
    Rows, Cols = best_layout(Tot)
    Position = range(1, Tot + 1)
    n = 0

    # fig = plt.figure(1, constrained_layout=True)
    for experiment in unique_experiments:
        # add every single subplot to the figure with a for loop
        ax = fig.add_subplot(Rows, Cols, Position[n])
        adapted_fontsize = 10 - int(Cols) + 2
        font_settings["fontsize"] = adapted_fontsize
        LengthDistributionSingle(
            sequencing_report,
            sample=experiment,
            region_of_interest=region_string,
            ax=ax,
            font_settings=font_settings,
            title_type="multi",
        )
        n += 1
    title = "\n".join(wrap("Length Distribution of given Samples", 40))
    fig.suptitle(title)


#   plt.show()
# plt.bar(unique_length, counts_length)
# p_values = counts_length/np.sum(counts_length)
# normalized on one:
# plt.bar(unique_length, counts_length/np.max(counts_length))
# p-values
#  plt.bar(unique_length, p_values)



from argparse import ArgumentParser
import os
import pandas as pd


class TestArgs:
    def __init__(self, args):
        self.args = args
        if hasattr(self.args, "sequencing_report"):
            self.report = self.check_sequencing_report()
        if hasattr(self.args, "region_plots"):
            self.region_of_interest = self.check_region_plots()

    def check_sequencing_report(self):
        assert self.args.sequencing_report.endswith(
            ".csv"
        ), "Your filename does not end with csv"
        assert os.path.isfile(
            self.args.sequencing_report
        ), "Could not find the path to the sequencing report file."
        report = pd.read_csv(self.args.sequencing_report)
        assert isinstance(
            report, pd.DataFrame
        ), "Could not read sequencing report. Please verify that your input file is a csv file."
        return report

    def check_tsv_dir(self):
        assert (
            os.path.isdir(self.args.tsv_dir) == True
        ), "Path to directory with tsv files does not exist"

    def check_region(self):
        assert self.args.region in [
            "targetSequences",
            "FR1",
            "CDR1",
            "FR2",
            "CDR2",
            "FR3",
            "CDR3",
            "FR4",
        ]

    def check_length_threshold(self):
        assert self.args.length_threshold >= 0, "Length threshold cannot be negative"

    def check_min_read_count(self):
        assert self.args.min_read_count >= 0, "Minimal read count cannot be negative"

    def check_remove_gaps(self):
        assert type(self.args.remove_gaps) == bool, "--remove_gaps must be boolean"

    def check_remove_errors(self):
        assert type(self.args.remove_errors) == bool, "--remove_errors must be boolean"

    def check_save_csv(self):
        assert (
            self.args.save_csv.endswith(".csv") == True
        ), "Your output filename must end with .csv"
        dir_csv = os.path.dirname(self.args.save_csv)
        assert os.path.isdir(
            dir_csv
        ), f"The directory for {os.path.basename(self.args.save_csv)} does not exist"

    def check_samples(self):
        assert len(self.args.samples) >= 1, "Please insert a sample name"
        for sample in self.args.samples:
            assert (
                sample in self.report["Experiment"].unique()
            ), f"Could not find {sample} in sequnecing_report"

    def check_single_sample(self):
        assert type(self.args.single_sample) == str, "Please insert a sample name"

        assert (
            self.args.single_sample in self.report["Experiment"].unique()
        ), f"Could not find {self.args.single_sample} in sequnecing_report"

    def check_pca_components(self):
        if self.args.batch_size:
            assert (
                self.args.pca_components < self.args.batch_size
            ), "Your number of pca components must be larger than your batch size"
        assert type(self.args.pca_components) == int
        assert (
            self.args.pca_components > 2
        ), "You must insert more than two pca components"

    def check_perplexity(self):
        assert self.args.perplexity >= 2, "Please enter more than 2 for perplexity"
        if self.args.batch_size:
            assert (
                self.args.perplexity < self.args.batch_size
            ), "Your batch size must be larger than the perplexity value"

    def check_iterations_tsne(self):
        assert (
            self.args.iterations_tsne > 251
        ), "Please enter more than 251 for the number of the iterations in tsne"

    def check_model_type(self):
        assert self.args.model_type in [
            "facebook/esm2_t6_8M_UR50D",
            "Rostlab/ProstT5_fp16",
            "Rostlab/prot_t5_xl_uniref50",
            "Rostlab/prot_t5_base_mt_uniref50",
            "Rostlab/prot_bert_bfd_membrane",
            "Rostlab/prot_t5_xxl_uniref50",
            "Rostlab/ProstT5",
            "Rostlab/prot_t5_xl_half_uniref50-enc",
            "Rostlab/prot_bert_bfd_ss3",
            "Rostlab/prot_bert_bfd_localization",
            "Rostlab/prot_t5_xl_bfd",
            "Rostlab/prot_bert",
            "Rostlab/prot_xlnet",
            "Rostlab/prot_bert_bfd",
            "Rostlab/prot_t5_xxl_bfd",
        ], "please enter a valid model"

    def check_batch_size(self):
        assert type(self.args.batch_size) == int

    def check_binding_data(self):
        binding_data = self.args.binding_data
        if binding_data != None:
            assert binding_data.endswith("csv")
            df_bind = pd.read_csv(binding_data)
            bind_cols = df_bind.columns.to_list()
            assert (
                self.region_of_interest in bind_cols
            ), "Your region you want to analyse must exist in your binding data"
            assert (
                len(bind_cols) >= 2
            ), "Besides your sequences you must also have values."
            assert df_bind.shape[0] >= 1, "you must have at least one value"

    def check_region_plots(self):
        region_of_interest = self.args.region_plots
        regions = [
            "aaSeqtargetSequences",
            "aaSeqFR1",
            "aaSeqCDR1",
            "aaSeqFR2",
            "aaSeqCDR2",
            "aaSeqFR3",
            "aaSeqCDR3",
            "aaSeqFR4",
        ]
        assert (
            region_of_interest in regions
        ), f"Please enter a valid region of interest from {regions}"
        assert (
            region_of_interest in self.report.columns.to_list()
        ), "The region you would like to analyse does not exist in the current sequencing report. Please change that in the highest level."
        return region_of_interest

    def check_antigen_names(self):
        if self.args.antigen_names != None:
            assert (
                self.args.binding_data != None
            ), "You cannot parse a value here without having binding data."
            df_bind = pd.read_csv(self.args.binding_data)
            for antigen in self.args.antigen_names:
                assert (
                    antigen in df_bind.columns.to_list()
                ), f"{antigen} is not in your binding data"

    def check_method_diversity(self):
        assert self.args.method in [
            "Shannon",
            "InverseSimpson",
        ], "please enter a one of the following values: [Shanon, InverseSimpson] "

    def check_n_neighbors(self):
        assert self.args.n_neighbors >= 1, "Please enter a value larger than 1"
        assert self.args.n_neighbors < self.args.batch_size * len(
            self.args.samples
        ), "Your number of neighbors must be smaller than your batch size"

    def check_min_dist(self):
        assert self.args.min_dist > 0, "Please enter a value larger than 0"

    def check_metric(self):
        avail_metrics = [
            "euclidean",
            "manhatten",
            "chebyshev",
            "minkowski",
            "canberra",
            "braycurtis",
            "haversine",
            "mahalanobis",
            "wminkowski",
            "seuclidean",
            "cosine",
            "correlation",
        ]
        assert (
            self.args.metric in avail_metrics
        ), f"Please enter a metric in {avail_metrics}"

    def check_characteristic(self):
        avail_characteristic = [
            "isoelecrtric_point",
            "aliphatic_index",
            "hydrophobicity",
            "weight",
            "mass_charge_ratio",
            "length",
            None,
            "binding",
        ]
        assert (
            self.args.characteristic in avail_characteristic
        ), f"Please enter a valid characteristic in {avail_characteristic}"

    def check_eps_dbscan(self):
        assert self.args.eps > 0, "Please enter a value larger than 0"
        assert type(self.args.eps) == float, "Please enter a float value"

    def check_min_pts_dbscan(self):
        assert self.args.min_pts >= 2, "Please enter a value larger than 2"
        assert type(self.args.min_pts) == int, "Please enter an integer value"

    def check_point_size(self):
        assert self.args.point_size > 5, "Please enter a value larger than 5"
        assert type(self.args.point_size) == int, "Please enter an integer value"

    def check_n_jobs(self):
        assert type(self.args.n_jobs) == int, "Please enter an integer value"
        assert self.args.n_jobs >= -1, "Please enter a value larger than -1"

    def check_chosen_seq_length(self):
        assert type(self.args.chosen_seq_length) == int, "Please enter an integer value"
        assert self.args.chosen_seq_length >= 1, "Please enter a value larger than 1"
        
    def check_method_logo(self):
        assert self.args.method_logo in ['bits', '', "proportional"], "Please enter a valid method for the logo plot"

    def check_adjusted_clone_size(self):
        assert self.args.clone_size_adjustment > 0, "Please enter a value larger than 0"
        assert type(self.args.clone_size_adjustment) == int, "Please enter an integer value"
        
    def check_embedding_vector_path(self):
        dir = os.path.dirname(self.args.embedding_vector_path)
        assert os.path.isdir(dir), "The directory for the embedding vector path does not exist"
        
    def check_matrix_type(self):
        assert self.args.matrix_type in ['morosita_horn', 'sorensen', 'jaccard'], "Please enter a valid matrix type"
        
    def check_levenshtein_distance(self):
        assert self.args.levenshtein_distance >= 1, "Please enter a value larger than 1. Otherwise the plot does not make sense"
        assert type(self.args.levenshtein_distance) == int, "Please enter an integer value"
        
    def check_dimension(self):
        assert self.args.dimension in [2, 3], "Please enter a valid dimension"
        assert type(self.args.dimension) == int, "Please enter an integer value"
        

class ExpoSeqArgs:
    def __init__(self, **kwargs) -> None:
        self.parser = ArgumentParser(
            description="Argument parser for plot function", allow_abbrev=True, **kwargs
        )
        self.chosen_tests = []

    def add_save_csv(self, flag="--save_csv"):
        self.parser.add_argument(
            flag,
            help="This must be parsed as path to tell where the output should be saved.",
            required=True,
        )
        self.chosen_tests.append("check_save_csv")

    def add_sequencing_report(self, shortcut="-r", flag="--sequencing_report"):
        self.parser.add_argument(
            shortcut,
            flag,
            help="Insert the path to the sequencing report (merged csv file of all samples) here.",
            required=True,
        )
        self.chosen_tests.append("check_sequencing_report")

    def add_samples(self, shortcut="-s", flag="--samples", required=True, default = None):
        self.parser.add_argument(
            shortcut,
            flag,
            help="Add a list of sample names here.",
            required=required,
            type=str,
            nargs="+",
            default = default
        )
        if required:
            self.chosen_tests.append("check_samples")

    def add_single_sample(self, flag="--single_sample"):
        self.parser.add_argument(
            flag,
            help="Add a string for the sample name you want to analyse.",
            required=True,
            type=str,
        )
        self.chosen_tests.append("check_single_sample")

    def add_region_of_interest(self, flag="--region", default="CDR3"):
        self.parser.add_argument(
            flag,
            help="This must be given to indicate which region should be analysed.",
            required=True,
            type=str,
        )
        self.chosen_tests.append("check_region")

    def add_region_plots(self, flag="--region_plots", default="aaSeqCDR3"):
        self.parser.add_argument(
            flag,
            help="This parses the region of interest starting with aaSeq to the plot.",
            type=str,
            required=True,
        )
        self.chosen_tests.append("check_region_plots")

    def add_pca_components(self, flag="--pca_components", default_value=50):
        self.parser.add_argument(
            flag,
            help="Add an integer for the number of pca_components you want to reduce the embedding to",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append("check_pca_components")

    def add_perplexity(self, flag="--perplexity", default_value=20):
        self.parser.add_argument(
            flag,
            help="Add an integer for the perplexity value the t-SNE algorithm needs to arrange your points in a certain cluster size.",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append("check_perplexity")

    def add_iterations_tsne(self, flag="--iterations_tsne", default_value=1500):
        self.parser.add_argument(
            flag,
            help="Add an integer for the number of iterations the t-SNE algorithm should take to arrange the coordiantes for the points.",
            default=default_value,
            type=int,
        )
        self.chosen_tests.append("check_iterations_tsne")

    def add_batch_size(self, flag="--batch_size", default_value=600):
        self.parser.add_argument(
            flag,
            help="Add an integer for the batch size you want to take per sample which equals to the number of sequences with the top clone counts",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append("check_batch_size")

    def add_model_type(
        self, flag="--model_type", default_value="Rostlab/prot_t5_xl_half_uniref50-enc"
    ):
        self.parser.add_argument(
            flag,
            help="Add a str as model type from the available models to embed your sequences in a high dimensional space.",
            type=str,
            default=default_value,
        )
        self.chosen_tests.append("check_model_type")

    def add_strands(self, flag="--show_strands", default_value=True):
        self.parser.add_argument(
            flag,
            help="Enter a boolean to show a batch of sequences in your plots as identifier. False does not show any.",
            type=bool,
            default=default_value,
        )

    def add_tsv_dir(self, flag="--tsv_dir"):
        self.parser.add_argument(
            flag,
            help="Add a path to a directory with tsv files here to generate the sequencing report.",
            type=str,
            required=True,
        )
        self.chosen_tests.append("check_tsv_dir")

    def add_length_threshold(self, flag="--length_threshold", default_value=5):
        self.parser.add_argument(
            flag,
            help="Enter the threshold for the minimal sequence length in the sequencing report here.",
            default=default_value,
            type=int,
        )
        self.chosen_tests.append("check_length_threshold")

    def add_min_read_count_threshold(self, flag="--min_read_count", default_value=3):
        self.parser.add_argument(
            flag,
            help="Add a value for the minimum required read count certain sequences should have to remain in the sequencing report.",
            default=default_value,
            type=int,
        )
        self.chosen_tests.append("check_min_read_count")

    def add_remove_gaps(self, flag="--remove_gaps", default_value=False):
        self.parser.add_argument(
            flag,
            help="Here you can enter whether you would like to remove gaps in your sequencing data. That means that your target sequence should be divisible by 3.",
            default=default_value,
            type=bool,
        )
        self.chosen_tests.append("check_remove_gaps")

    def add_remove_errors(self, flag="--remove_errors", default_value=True):
        self.parser.add_argument(
            flag,
            help="If set to True this will remove all sequences (rows) in the sequencing report which have a *.",
            default=default_value,
            type=bool,
        )
        self.chosen_tests.append("check_remove_errors")

    def add_binding_data(self, flag="--binding_data", default_value=None):
        self.parser.add_argument(
            flag,
            help="Add the path to the csv file with the binding data here, and it will be included in the plot.",
            type=str,
            default=default_value,
        )
        self.chosen_tests.append("check_binding_data")

    def add_antigen_names(self, flag="--antigen_names", default_value=None):
        self.parser.add_argument(
            flag,
            help="Your antigen names must be the column names of your binding data. You can insert multiple and they must be a string. You cannot parse names without having binding data.",
            nargs="+",
            type=str,
            default=default_value,
        )
        self.chosen_tests.append("check_antigen_names")

    def method_diversity(self, flag="--method"):
        self.parser.add_argument(
            flag,
            help="Choose the method to calculate the diversity in your samples. Default is Shannon for the ShannonIndex",
            type=str,
            default="Shannon",
        )
        self.chosen_tests.append("check_method_diversity")

    def add_n_neighbors(self, flag="--n_neighbors", default_value=15):
        self.parser.add_argument(
            flag,
            help="Larger values will result in more global structure being preserved at the loss of detailed local structure. In general this parameter should be between 5 to 50.",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append("check_n_neighbors")

    def add_min_dist(self, flag="--min_dist", default_value=0.05):
        self.parser.add_argument(
            flag,
            help="This parameter controls how tightly you want to compress the clusters together. Larger values will result in more compact clusters.",
            type=float,
            default=default_value,
        )
        self.chosen_tests.append("check_min_dist")

    def add_metric(self, flag="--metric", default_value="euclidean"):
        self.parser.add_argument(
            flag,
            help="The metric to use to compute distances in the high dimensional space. Default is euclidean",
            type=str,
            default=default_value,
        )
        self.chosen_tests.append("check_metric")

    def add_characteristic(self, flag="--characteristic", default_value=None):
        self.parser.add_argument(
            flag,
            help="The characteristic you want to analyse. Default is length",
            type=str,
            default=default_value,
        )
        self.chosen_tests.append("check_characteristic")

    def add_eps_dbscan(self, flag="--eps", default_value=0.5):
        self.parser.add_argument(
            flag,
            help="Maximum distance between two points to still form one cluster",
            type=float,
            default=default_value,
        )
        self.chosen_tests.append("check_eps_dbscan")

    def add_min_pts_dbscan(self, flag="--min_pts", default_value=2):
        self.parser.add_argument(
            flag,
            help="fewest number of points required to form a cluster. 2 is seems to be quiet good.",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append("check_min_pts_dbscan")

    def add_point_size(self, flag="--point_size", default_value=300):
        self.parser.add_argument(
            flag,
            help="The moralized size of the points in the plot.",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append("check_point_size")

    def add_n_jobs(self, flag="--n_jobs", default_value=-1):
        self.parser.add_argument(
            flag,
            help="The number of jobs to run in parallel. -1 means all processors.",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append("check_n_jobs")
        
    def add_chosen_seq_length(self, flag="--chosen_seq_length", default_value=6):
        self.parser.add_argument(
            flag,
            help="The sequence length you want to analyse. If you do not insert a value the most frequent sequence length will be chosen.",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append('check_chosen_seq_length')

    def add_method_logo(self, flag="--method_logo", default_value=""):
        self.parser.add_argument(
            flag,
            help="The method to calculate the logo plot. Default is proportional.",
            type=str,
            default=default_value,
        )
        self.chosen_tests.append("check_method_logo")
        
    def add_embedding_vector_path(self, flag="--embedding_vector_path", default_value="temp"):
        self.parser.add_argument(
            flag,
            help="The path to the embedding vector path.",
            type=str,
            default=default_value,
        )
        self.chosen_tests.append("check_embedding_vector_path")
            
    def add_matrix_type(self, flag="--matrix_type", default_value="morosita_horn"):
        self.parser.add_argument(
            flag,
            help="The type of matrix you want to calculate. Default is morosita_horn",
            type=str,
            required = True,
            default=default_value,
        )
        self.chosen_tests.append("check_matrix_type")
    
    def add_levenshtein_distance(self, flag = "--levenshtein_distance", default_value=2):
        self.parser.add_argument(
            flag,
            help="The levenshtein distance you want to base the Analysis on.",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append("check_levenshtein_distance")
    
    def add_dimension(self, flag = "--dimension", default_value=2):
        self.parser.add_argument(
            flag,
            help="The dimension you want to base the Analysis on.",
            type=int,
            default=default_value,
        )
        self.chosen_tests.append("check_dimension")

def prep_args(args):
    parser = args.parser.parse_args()
    tests = args.chosen_tests
    TestInput = TestArgs(parser)
    for (
        test
    ) in (
        tests
    ):  # tests basically input values for chosen arguments. they should be all tested in the pipeline but this just prevents wrong inputs on a lower level.
        method = getattr(TestInput, test)
        method()
    return parser



import pandas as pd
import os
from tabulate import tabulate

class ChooseMethod:
    def __init__(self) -> None:
        preset_path = os.path.join(os.getcwd(), "settings", "preset_list.csv")
        self.open_presets(preset_path)
        self.print_table()
        

        
        
    def open_presets(self, preset_path):
        
        if os.path.isfile(preset_path):
            self.preset_table = pd.read_csv(preset_path, sep=";")
        else:
            FileNotFoundError
            
    def print_table(self):
        print(tabulate(self.preset_table, headers='keys', tablefmt='psql'))      
    
    def ask_user(self):
        avail_preset = self.preset_table["Preset"].tolist()
        avail_preset.append("")
        while True:
            use_method = input(
                "Per default you will align your data with the following method: milab-human-tcr-dna-multiplex-cdr3.\nPress enter if you want to continue.\nOtherwise type in the method of your choice.\nIt has to be the exact same string which is given on the Mixcr documentation.")    
            if use_method in avail_preset:
                if use_method == "":
                    use_method = "milab-human-tcr-dna-multiplex-cdr3"
                else:
                    pass
                break
            else:
                print("The method you entered is not in the table. Please enter a valid method.")
        return use_method
    
    def check_args(self, chosen_method):
        args_table = self.preset_table[self.preset_table["Preset"] == chosen_method]
        args_table = args_table.iloc[:, 2:]
        add_arguments = {}
        method_vals = args_table.values.tolist()[0]
        gen_columns = args_table.columns.tolist()
        for vals in method_vals: 
            if vals != str(0) and vals != int(0):
                index = method_vals.index(vals)
                add_arguments[gen_columns[index]] = vals
            else:
                pass
        return add_arguments
    
    @staticmethod
    def change_args(add_arguments):
        if len(add_arguments) >= 1:
            for arg, value in add_arguments.items():
                usr_input = input(f"Do you want to enter another value for {arg}.\nThe default value is {value}. Press enter to continue, but if you want to change the input enter the corresponding value.")
                if usr_input == "":
                    continue
                else:
                    add_arguments[arg] = usr_input
        else:
            pass
        return add_arguments
    
    

class Test:
    def __init__(self) -> None:
        Test_Choose = ChooseMethod()
        assert Test_Choose.check_args("milab-human-tcr-dna-multiplex-cdr3") == {}, "output must be an empty dict"
        assert Test_Choose.check_args("ont-rna-seq-vdj-full-length") == {"species": "hsa"}, "output must contain the hsa value with species key"
        assert Test_Choose.change_args({}) == {}, "No prompting should appear"
        Test_Choose.open_presets("") == FileNotFoundError, "An error should appear"
    



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from textwrap import wrap
import ExpoSeq.plots.matrices.morosita_horn_matrix as morosita_horn_matrix
import ExpoSeq.plots.matrices.sorensen_matrix as sorensen_matrix
import ExpoSeq.plots.matrices.jaccard_matrix as jaccard_matrix
import ExpoSeq.plots.matrices.relative_matrix as relative_matrix

class IdentityMatrix:
    def __init__(self, sequencing_report, region_string,matrix_type, colorbar_settings, specific_experiments = False,ax = None, protein = True, font_settings = {}, 
                 cmap = 'inferno', fmt = ".2f", annotate_cells = False):
        self.ax = ax

        self.matrix, self.unique_experiments = self.tidy_data(matrix_type, specific_experiments, sequencing_report, region_string, protein)

        self.font_settings = font_settings
        if ax != None:
            self.createPlot(cmap = cmap, fmt = fmt, annotate_cells = annotate_cells, colorbar_settings=colorbar_settings)
            if font_settings != {}:
                self.add_style()
                
    @staticmethod
    def tidy_data(matrix_type, specific_experiments, sequencing_report, region_string, protein):
        if matrix_type == "morosita_horn":
            matrix, unique_experiments = morosita_horn_matrix.PrepareData().cleaningPlot(specific_experiments, sequencing_report, protein, region_string)
        if matrix_type == "sorensen":
            matrix, unique_experiments = sorensen_matrix.PrepareData().sorensen_matrix(sequencing_report, protein, region_string, specific_experiments)        
        if matrix_type == "jaccard":
            matrix, unique_experiments = jaccard_matrix.PrepareData().cleaning_jaccard(sequencing_report, region_string, protein, specific_experiments)
        if matrix_type == "relative":
            matrix, unique_experiments = relative_matrix.PrepareData().heatmap_share(sequencing_report, region_string, protein, specific_experiments)
        
        return matrix, unique_experiments

    def createPlot(self,linewidth = 0,line_color = "white", annotate_cells = False, annot_kws = 6, fmt = ".2f", cmap = 'inferno', colorbar_settings = None):
        annot_kws = {"size": annot_kws}
        matrix = self.matrix.sort_index(axis=1)
        matrix = matrix.sort_index(axis = 0)
        if colorbar_settings != None:
            colorbar_sets = {**colorbar_settings,
                            **{'label': "Degree of Identity"}}
            self.morosita_horn_matrix = sns.heatmap(matrix,
                                        cbar_kws=colorbar_sets,
                                        ax = self.ax,
                                        linewidths=linewidth,
                                        linecolor = line_color,
                                        annot = annotate_cells,
                                        annot_kws=annot_kws,
                                        fmt = fmt,
                                        cmap= cmap)
        else:
            self.morosita_horn_matrix = sns.heatmap(matrix,
                            ax = self.ax,
                            linewidths=linewidth,
                            linecolor = line_color,
                            annot = annotate_cells,
                            annot_kws=annot_kws,
                            fmt = fmt,
                            cmap= cmap)


    def add_style(self):
        plt.xticks(ticks = np.arange(0.5, len(self.unique_experiments) + 0.5 ,1),
                labels = self.unique_experiments,
                rotation = 45,
                ha = 'right',
                size = 12) # create a function which finds the perfect size based on counts of xlabels
        plt.yticks(ticks = np.arange(0.5, len(self.unique_experiments) + 0.5 ,1),
                labels = self.unique_experiments,
                rotation = 0,
                va='center',
                size=12)
        original_fontsize = self.font_settings["fontsize"]
        self.font_settings["fontsize"] = 22
        title = "Identity between samples based on Morosita Horn index"
        title = "\n".join(wrap(title, 40))
        self.ax.set_title(title,pad = 12, **self.font_settings)
        self.font_settings["fontsize"] = original_fontsize



import seaborn as sns
import matplotlib.pyplot as plt
from textwrap import wrap


class LengthPlotMultiple:
    def __init__(
        self,
        sequencing_report,
        region_of_interest,
        ax=None,
        font_settings={},
        plot_type="boxplot",
    ):
        self.ax = ax

        sequencing_report = self.tidy(sequencing_report, region_of_interest)
        if ax != None:
            if plot_type == "boxplot":
                self.make_boxplot(sequencing_report)
            else:
                self.make_violin_plot(sequencing_report)
        if font_settings != {}:
            self.add_labels(font_settings)
            self.title(font_settings)

    @staticmethod
    def tidy(sequencing_report, region_of_interest):
        region = region_of_interest.replace("aaSeq", "nSeq")
        length_region = sequencing_report[region].str.len()
        sequencing_report["length_region"] = length_region
        return sequencing_report

    def make_boxplot(self, sequencing_report):
        sns.boxplot(
            x="length_region",
            y="Experiment",
            data=sequencing_report,
            palette="Dark2",
            ax=self.ax,
        )
        sns.despine(top=True, right=True, bottom=True, left=True, ax=self.ax)

    def make_violin_plot(self, sequencing_report):
        sns.violinplot(
            sequencing_report,
            x="length_region",
            y="Experiment",
            inner="box",
            palette="Dark2",
            ax=self.ax,
        )
        sns.despine(top=True, right=True, bottom=True, left=True, ax=self.ax)

    def add_labels(self, font_settings):
        self.ax.set_ylabel("Read Count", **font_settings)  # Y label
        self.ax.set_xlabel("Read Length", **font_settings)  # X label

    @staticmethod
    def title(font_settings):
        original_fontsize = font_settings["fontsize"]
        font_settings["fontsize"] = 20
        title = "\n".join(wrap("Length Distribution of all samples ", 40))
        plt.title(title, pad=12, **font_settings)
        font_settings["fontsize"] = original_fontsize



import os
import pandas as pd
import glob
import markdown
from .layout_finder import best_layout

class QuartoBuilder:
    def __init__(self, title, figure = False):
        self.content = ""
        self.title = title
        self.create_header(figure)

    def write_quarto(self, save_dir):
        save_path = os.path.join(save_dir,f"{self.title}.qmd" )
        with open(save_path, "w") as f:
            f.write(self.content)
            

    def create_header(self, figure):
        self.content += "---\n"
        self.content += f"title: {self.title}\n"
        self.content += f"format:"
        self.content += "\n     html:"
        self.content += "\n         self-contained: true"
        if figure == False:
            self.content += "\n         toc: true"
            self.content += "\n         toc-location: left"
            self.content += "\n         toc-expand: 2"
        else:
            self.content += "\n         page-layout: full"
            

        self.content += "\n"
        self.content += "---"

    def add_subplot_figures(self, files, captures):
        rows, cols = best_layout(len(files))
        self.next_row()
        inside_string = "{layout-ncol=" + str(cols) + "}"
        self.content += f"::: {inside_string}"
        self.next_row()
        if len(captures) != len(files):
            raise SystemError
        for picture , capture in zip(files, captures) :
            self.content += f"![{capture}]"
            self.content += f"({picture})"
            self.next_row()
        self.content += ":::"
        
    
        
    def end_python_code(self):
        self.next_row()
        self.content += "```"
        
    
    
    def create_headline(self, header, section = "#"):
        self.next_row()
        self.content += section + " " + header
    
    def next_row(self):
        self.content += "\n\n"
        
    def add_page(self):
        self.next_row()
        self.content += "{{< pagebreak >}}"
        self.next_row()
        
    def add_text(self, text):
        self.next_row()
        self.content += text
        

    def figure_text(self, figure_path, text,figure_description, position_text = "left"):
        self.next_row()
        self.content += ":::"
        self.content += "{layout-ncol=" + str(2) + "}"
        self.next_row()
        if position_text == "left":
            self.content += f"![]"
            self.content += f"({figure_path})"
            self.next_row()
            self.content += text
        else:
            self.content += text
            self.next_row()
            self.content += f"![]"
            self.content += f"({figure_path})"
            self.next_row()
            self.content += figure_description
            self.next_row()
            self.content += ":::"
    
                
        
    def add_figure_horizontal(self, ncol, picture_paths: list, figure_description, figure_subtitles = None):
        self.next_row()
        self.content += ":::"
        self.content += " {layout-ncol=" + str(ncol) + "}"
        self.next_row() 
        for index_pic in range(len(picture_paths)):
            if figure_subtitles != None:
                self.content += f"![{figure_subtitles[index_pic]}]"
            else:
                self.content += f"![]"
            self.content += f"({picture_paths[index_pic]})"
           # self.content += "{" + f"width={str(figure_width)}" + "%}"
            self.next_row()
        self.content += figure_description
        self.next_row()
        self.content += ":::"
        
    def add_figure_layout(self, layout_option, picture_paths: list):
        self.next_row()
        self.content += ":::"
        self.content += ' {layout=' + f'"{layout_option}"' + "}"
        self.next_row()
        for index_pic in range(len(picture_paths)):
            self.content += f"![]"
            self.content += f"({picture_paths[index_pic]})"
           # self.content += "{" + f"width={str(figure_width)}" + "%}"
            self.next_row()
        self.content += ":::"
        
    def add_figure(self, picture_path):
        self.next_row()
        self.content += ":::"
        self.content += " {layout-ncol=" + str(1) + "}"
        self.next_row()
        self.content += f"![]"
        self.content += f"({picture_path})"
        self.next_row()
        self.content += ":::"
        
    def md_to_text(self, md_file):
        self.next_row()
        with open(md_file, "r") as f:
            markdown_text = f.read()
        text = markdown.markdown(markdown_text)
        self.content += text
        
    def add_table(self, filename, top_,cols = None, max_column = None, ):
        self.next_row()
        df = pd.read_excel(filename)

        # Get the headers
        if cols != None:
            df = df[cols]
        else:
            pass
        
        if top_ != None:
            df = df.head(top_)
        else:
            df = df.nlargest(10, max_column)
            
        heads_df = df.columns.tolist()
        self.content += f"{df.to_markdown()}"
        
        

      #  headers = df.columns.tolist()

        # Convert headers to markdown format
      #  markdown_headers = ' | '.join(headers)
      #  markdown_divider = ' | '.join(['---'] * len(headers))
      #  self.content += f"{markdown_headers}\n{markdown_divider}"
        
        
        

def check_path_multiple(files:list):
    for i in files:
        if i == None:
            return False
        if os.path.isfile(i):
            pass
        else:
            return False
    return True

def find_file_with_substring(directory, substring):
    for root, dirs, files in os.walk(directory):
        for file in files:
            if substring in file:
                return os.path.join(root, file)
    return None

def find_multiple_files_with_substring(directory, substring):
    if substring == None:
        return None
    matching_files = []
    all_files = glob.glob(directory + "/*")
    for file in all_files:
        if substring in file:
            matching_files.append(os.path.join(directory, file))
    return matching_files
    

    


def create_quarto(experiment, plot_path, binding_data, samples):
    assert os.path.isdir(plot_path), f"The directory {plot_path} does not exist"
    assert os.path.isdir(os.path.join(plot_path, "sequence_cluster")), f"The directory {os.path.join(plot_path, 'sequence_cluster')} does not exist"
    assert os.path.isdir(os.path.join(plot_path, "sequence_embedding")), f"The directory {os.path.join(plot_path, 'sequence_embedding')} does not exist"
    assert os.path.isdir(os.path.join(plot_path, "sequence_cluster", "reports")), f"The directory {os.path.join(plot_path, 'sequence_cluster', 'reports')} does not exist"
    assert os.path.isdir(os.path.join(plot_path, "length_distributions")), f"The directory {os.path.join(plot_path, 'length_distributions')} does not exist"
    assert os.path.isdir(os.path.join(plot_path, "rarefraction_curves")), f"The directory {os.path.join(plot_path, 'rarefraction_curves')} does not exist"
    assert os.path.isdir(os.path.join(plot_path, "logo_plots")), f"The directory {os.path.join(plot_path, 'logo_plots')} does not exist"
    Builder = QuartoBuilder(experiment)
    region = os.path.basename(plot_path)
    Builder.create_headline(f"Basic Overview - {region}")
    Builder.create_headline("Theory", section = "##")
    if os.path.isfile(os.path.join("settings", "rarefraction_curves_desc.md")):
        Builder.md_to_text(os.path.join("settings", "rarefraction_curves_desc.md"))
    if os.path.isfile(os.path.join("settings", "alignment_plot.md")):
        Builder.md_to_text(os.path.join("settings", "alignment_plot.md"))
    if os.path.isfile(os.path.join("settings", "morosita_horn.md")):
        Builder.md_to_text(os.path.join("settings", "morosita_horn.md"))
    if os.path.isfile(os.path.join("settings", "jaccard.md")):
        Builder.md_to_text(os.path.join("settings", "logo_plot.md"))

    #Builder.
    Builder.create_headline("Identity between samples", section = "##")
    Builder.create_headline("Identity based on Morosita Horn Index", section = "###")
    morosita_horn_path = find_file_with_substring(plot_path, "morosita_horn")
    if morosita_horn_path != None:
        Builder.add_figure(morosita_horn_path)
    else: pass
    Builder.create_headline("Identity based on Jaccard Index", section = "###")
    jaccard_path = find_file_with_substring(plot_path, "jaccard")
    if jaccard_path != None:
        Builder.add_figure(jaccard_path)
    else: pass
    Builder.add_page()
    Builder.create_headline("Identity based on Sorensen Index", section = "###")
    sorensen_path = find_file_with_substring(plot_path, "sorensen")
    Builder.add_page()
    if sorensen_path != None:
        Builder.add_figure(sorensen_path)
    else: pass
    Builder.add_page()
    
    Builder.create_headline("Sequencing Quality of samples", section = "##")
    figure_alignment = os.path.join(plot_path, "alignment_quality.png")
    figure_rarefraction = os.path.join(plot_path, "rarefraction_all.png")
    
    if check_path_multiple([figure_alignment, figure_rarefraction]) != False:
        Builder.add_figure_horizontal(2, [figure_alignment, figure_rarefraction], "")
    else:
        for i in [figure_alignment, figure_rarefraction]:
            if i != None:
                Builder.add_figure(i)
            else:
                pass
    Builder.add_page()
    # Diversity Plots
    figure_diversity_shannon = os.path.join(plot_path, "diversity_Shannon.png")
    figure_diversity_simp = os.path.join(plot_path, "diversity_InverseSimpson.png")
    if check_path_multiple([figure_diversity_shannon, figure_diversity_simp]) != False:
        Builder.add_figure_horizontal(2, [figure_diversity_shannon, figure_diversity_simp], "")
    else:
        for i in [figure_diversity_shannon, figure_diversity_simp]:
            if i != None:
                Builder.add_figure(i)
            else:
                pass
    Builder.add_page()
    
    Builder.create_headline("Sample specific quality analysis", section = "##")
    length_distribution_path = os.path.join(plot_path, "length_distributions")
    rarefraction_curve_path = os.path.join(plot_path, "rarefraction_curves")
    clone_fraction_path = os.path.join(plot_path, "clone_fraction")
    
    for sample in samples:
        Builder.create_headline("Quality analysis for " + sample, section = "###")
        length_distr_single = find_file_with_substring(length_distribution_path, sample)
        rarefraction_single = find_file_with_substring(rarefraction_curve_path, sample)
        clone_single = find_file_with_substring(clone_fraction_path, sample)
        if check_path_multiple([length_distr_single, rarefraction_single, clone_single]) != False:
            Builder.add_figure_layout([[1, 1], [1]], [length_distr_single, rarefraction_single, clone_single])
        else:
            for i in [length_distr_single, rarefraction_single, clone_single]:
                if i != None:
                    Builder.add_figure(i)
                else: 
                    pass
        Builder.add_page()
        
    Builder.create_headline("Sequence Logo Plots", section = "##")
    logo_plot_path = os.path.join(plot_path, "logo_plots")
    for sample in samples:
        Builder.create_headline("Logo Plot for sequences with length 16 for " + sample, section = "###")
        logo_plot_single = find_file_with_substring(logo_plot_path, sample)
        if logo_plot_single != None:
            Builder.add_figure(logo_plot_single)
        else:
            pass
        Builder.add_page()

    ls_clusters_dir = os.path.join(plot_path, "sequence_cluster")
    Builder.create_headline("Sequence Clusters")
    Builder.create_headline("Theory", section = "##")
    if os.path.isfile(os.path.join("settings", "levenshtein_cluster.md")):
        Builder.md_to_text(os.path.join("settings", "levenshtein_cluster.md"))
    if os.path.isfile(os.path.join("settings", "dendrogram_cluster.md")):
        Builder.md_to_text(os.path.join("settings", "dendrogram_cluster.md"))
    if os.path.isfile(os.path.join("settings", "sequence_embedding.md")):
        Builder.md_to_text(os.path.join("settings", "sequence_embedding.md"))

    report_seq_cluster = os.path.join(ls_clusters_dir, "reports")
    
    Builder.create_headline("Levenshtein distance based", section="##")
    Builder.create_headline("Relation of all samples", section = "###")
    _isfile = find_file_with_substring(plot_path, "ls_connection_all")
    if _isfile != None:
        Builder.add_figure(_isfile)
    else:
        pass
    for sample in samples:
        
        plot_files = find_multiple_files_with_substring(ls_clusters_dir, sample)
        if check_path_multiple(plot_files):
            Builder.create_headline(f"Sequence Similarity based on Levenshtein Distance of {sample}", "###")
            Builder.add_figure_horizontal(ncol = 2, picture_paths=plot_files, figure_description="")
        else:
            Builder.create_headline(f"Sequence Similarity based on Levenshtein Distance of {sample}", "###")
            for i in plot_files:
                if i != None:
                    Builder.add_figure(i)
                else:
                    pass
        try:
            report_file = find_file_with_substring(report_seq_cluster, sample)
            if report_file != None:
                Builder.add_table(report_file, top_ = 10)
        except:
            pass

        Builder.add_page()

    Builder.create_headline("Sequence Similarity between samples", section = "##")
    embedding_dir = os.path.join(plot_path, "sequence_embedding", "sgt")
    Builder.create_headline("SGT Embedder", section = "###")
    for sample in samples:
        file = find_file_with_substring(embedding_dir, sample)
        if file != None:
            Builder.create_headline(f"Global Sequence similarity for {sample}", section = "####")
            Builder.add_figure(file)
            Builder.add_page()
        else:
            pass
        
    Builder.create_headline("ProtBert Embedder", section = "###")
    embedding_dir = os.path.join(plot_path, "sequence_embedding", "protbert")
    for sample in samples:
        file = find_file_with_substring(embedding_dir, sample)
        if file != None:
            Builder.create_headline(f"Global Sequence similarity for {sample}", section = "####")
            Builder.add_figure(file)
            Builder.add_page()
        else:
            pass
    
    Builder.create_headline("T5 Embedder", section = "###")
    embedding_dir = os.path.join(plot_path, "sequence_embedding", "T5")
    for sample in samples:
        file = find_file_with_substring(embedding_dir, sample)
        if file != None:
            Builder.create_headline(f"Global Sequence similarity for {sample}", section = "####")
            Builder.add_figure(file)
            Builder.add_page()
        else:
            pass
    
    Builder.write_quarto(save_dir=plot_path)
    
    if binding_data is not None:
        if os.path.basename(plot_path) in binding_data.columns.tolist():
            assert os.path.isdir(os.path.join(plot_path, "clustering_antigens")), f"The directory {os.path.join(plot_path, 'clustering_antigens')} does not exist"
            assert os.path.isdir(os.path.join(plot_path, "clustering_antigens", "ls_binding_cluster")), f"The directory {os.path.join(plot_path, 'clustering_antigens', 'ls_binding_cluster')} does not exist"
            assert os.path.isdir(os.path.join(plot_path, "clustering_antigens", "dendro_binding")), f"The directory {os.path.join(plot_path, 'clustering_antigens', 'dendro_binding')} does not exist"
            assert os.path.isdir(os.path.join(plot_path, "clustering_antigens", "reports")), f"The directory {os.path.join(plot_path, 'clustering_antigens', 'reports')} does not exist"
            Builder.create_headline("Cluster Antigens")
            Builder.create_headline("Theory", section = "##")
            if os.path.isfile(os.path.join("settings", "reports_binding.md")):
                Builder.md_to_text(os.path.join("settings", "reports_binding.md"))

            embedding_antigen_path = os.path.join(plot_path, "clustering_antigens")
            embedding_antigen_report = os.path.join(embedding_antigen_path, "reports")
            Builder.create_headline("Cluster with embedding", section = "##")
            for sample in samples:
                file = find_file_with_substring(embedding_antigen_path, sample)
                if file != None:
                    Builder.create_headline(f"Sequence embedding of {sample}", section = "###")
                    Builder.add_figure(file)
                    try:
                        report = find_file_with_substring(embedding_antigen_report, sample)
                        cols_choice = ["tsne1", "tsne2", "binding", "sequences", "sequence_id"]
                        cur_df = pd.read_excel(report)
                        df_cols = cur_df.columns.tolist()
                        incorrect_cols = [x for x in cols_choice if x not in df_cols]
                        assert not incorrect_cols, "Either the user changed the default column header or the developer for the tsne clusters with the antigen"
                        Builder.add_table(report, top_ = None, cols = cols_choice, max_column = "binding")
                    except:
                        print("Something went wrong with creating the tsne binding plots in the report")
                else:
                    pass
            ls_cluster_binding = os.path.join(plot_path,"clustering_antigens", "ls_binding_cluster")
            
            ls_cluster_binding_report = os.path.join(ls_cluster_binding, "reports")
            dendro_cluster_binding = os.path.join(plot_path,"clustering_antigens", "dendro_binding")
            Builder.create_headline("Cluster binding data based on levenshtein distance", section = "##")
            for sample in samples:
                ls_file = find_file_with_substring(ls_cluster_binding, sample)
                dendro_file = find_file_with_substring(dendro_cluster_binding, sample)
                if ls_file != None:
                    Builder.create_headline(f"LS-Distance cluster for {sample}", section = "###")
                    if dendro_file != None:
                        Builder.add_figure(ls_file)
                        Builder.add_figure(dendro_file)
                #      Builder.add_figure_horizontal(ncol = 2, picture_paths=[ls_file, dendro_file], figure_description="Cluster visualizations based on LS-distance")
                    else:
                        Builder.add_figure(ls_file)
                    try:
                        report_file = find_file_with_substring(ls_cluster_binding_report, sample)
                        Builder.add_table(report_file, top_ = 10)
                    except:
                        print("Could not create report table for ls cluster")
                else:
                    pass
        else:
            print(f"No plots with binding data were included in the report.\n If you have plots with binding data in {plot_path} please make sure that you upload the binding data to the pipeline with plot.add_binding_data() before you run plot.create_report()")
    Builder.write_quarto(save_dir=plot_path)




from scipy.cluster.hierarchy import dendrogram, linkage
from matplotlib import pyplot as plt
from scipy.spatial.distance import squareform
import pandas as pd
import warnings
from textwrap import wrap
import editdistance
import numpy as np


class PrepareData:
    
    @staticmethod
    def create_distance_matrix(aa, max_dist=20):  # max_dist is a large finite value
        num_sequences = len(aa)
        distance_matrix = np.full(
            (num_sequences, num_sequences), max_dist
        )  # initialize with max_dist

        for i in range(num_sequences):
            for j in range(i + 1, num_sequences):
                lev_distance = editdistance.distance(aa[i], aa[j])
                distance_matrix[i, j] = lev_distance
                distance_matrix[j, i] = (
                    lev_distance  # The distance matrix is symmetric
                )

        return distance_matrix

    @staticmethod
    def get_clustered_sequences(aa, max_cluster_dist):
        clustered_sequences = set()

        for i in range(len(aa)):
            for j in range(i + 1, len(aa)):
                if editdistance.distance(aa[i], aa[j]) < max_cluster_dist:
                    clustered_sequences.add(aa[i])
                    clustered_sequences.add(aa[j])

        return list(clustered_sequences)

    def tidy(
        self, sequencing_report:pd.DataFrame, samples:list, batch_size:int, region_string:str, max_cluster_dist:int
    ):
        """
        linked: linkage matrix which contains per sequence 4 values.
        aa_clustered: List of sequences which length is the same as the number of arrays in the linkage matrix.
        """
        
        sample_report = sequencing_report[
            sequencing_report["Experiment"].isin(samples)
        ]  ## insert test if sample not found
        report = sample_report.groupby("Experiment").head(batch_size)
        report = report.drop_duplicates(subset=[region_string])
        aa = list(report[region_string])
        aa_clustered = self.get_clustered_sequences(aa, max_cluster_dist)
        # Create the distance matrix using the filtered list of sequences
        levenshtein_distance_matrix = self.create_distance_matrix(aa_clustered)

            # Convert to condensed form and create the dendrogram
        condensed_matrix = squareform(levenshtein_distance_matrix, checks=False)
        linked = linkage(
            condensed_matrix, "single"
        )  # nearest point algorithm, n-1 matrix returned

        return linked, aa_clustered


class LevenshteinDend:
    def __init__(
        self,
        sequencing_report,
        region_of_interest,
        samples: list,
        font_settings={},
        batch_size=300,
        max_cluster_dist=2,
        ax=None,
    ) -> None:
        linked, aa_clustered = PrepareData().tidy(
            sequencing_report, samples, batch_size, region_of_interest, max_cluster_dist
        )
        if ax != None:
            self.plot(linked, ax, aa_clustered, font_settings, samples)

    def plot(self, linked, ax, aa_clustered, font_settings, samples):
        dendrogram(
            linked,
            orientation="right",
            distance_sort="descending",
            show_leaf_counts=True,
            labels=aa_clustered,
            ax=ax,
        )
        ax.set_xlabel("Levenshtein Distance", **font_settings)
        ax.set_ylabel("Sequences", **font_settings)
        title = "\n".join(
            wrap("Levenshtein Distance between sequences in " + " ".join(samples), 40)
        )
        ax.set_title(title, pad=12, **font_settings)
        plt.tight_layout()





class RUNMSA:
    def get_clustalw_path(self):
        global_params = self.Settings.read_global_vars()
        clustalw_path = global_params["clustalw_path"]
        try:
            import tkinter as tk
            from tkinter import filedialog
            if clustalw_path == '':
                print("Please choose the path to the clustalw exe with the filechooser. If you don't have it, you can download it from http://www.clustal.org/download/current/")
                clustalw_path = filedialog.askopenfilename()
                self.Settings.change_global_vars("clustalw_path", clustalw_path)
        except:
            print("Tk not installed. Please install it to use this functionality and visualize the MSA")
        return clustalw_path
    
    def run_MSA(self, batch_size, samples):
        path_clustalw = self.get_clustalw_path()
        if not path_clustalw == '' and os.path.isfile(path_clustalw):
            try:
                os.remove(self.outfile)
            except:
                pass
            sequences = self.get_top_sequences(batch_size, samples)
            self.create_fasta(sequences)
            clustalw_cline = ClustalwCommandline(path_clustalw, infile=self.seqs_path, outfile=self.outfile)
            stdout, stderr = clustalw_cline()
            os.remove(self.outfile)


import matplotlib.pyplot as plt
import random
from collections import Counter
from textwrap import wrap
import pandas as pd


class PrepareData:
    @staticmethod
    def cleaning_data(sequencing_report, samples, region_of_interest, fraction_column):
        sub_table = sequencing_report.loc[sequencing_report["Experiment"].isin(samples)]
        apply_fun = lambda x: list(x)
        sub_table = (
            sub_table.groupby("Experiment")[
                [region_of_interest, fraction_column, "readCount"]
            ]
            .agg(apply_fun)
            .reset_index()
        )
        return sub_table

    def tidy(
        self,
        sequencing_report,
        samples,
        region_of_interest,
        fraction_column="cloneFraction",
    ):
        region_of_interest = region_of_interest.replace("aaSeq", "nSeq")
        sub_table = self.cleaning_data(
            sequencing_report, samples, region_of_interest, fraction_column
        )
        x_axis = []
        y_axis = []
        sample_names = []
        for sample in range(len(sub_table)):
            sequences = list(sub_table.iloc[int(sample)][region_of_interest])
            fraction = list(sub_table.iloc[int(sample)][fraction_column])
            counter = 0
            temp_list = []
            unique_list = []
            total_list = []
            mean_count = sum(sub_table.iloc[sample]["readCount"]) / 100
            while counter < 100:
                temp = random.choices(
                    sequences, weights=fraction, k=int(mean_count)
                )  # randomly sample 100th of the sequences with the recalculated clonefractions as probablities
                temp_list.extend(
                    temp
                )  ## extend the temp_list with the samples from each 'pull-out' of the 100 'pull-outs'
                unique = Counter(
                    temp_list
                ).keys()  ## count unique seequences from the increasing temp_list
                unique_list.append(
                    unique
                )  ## for each run append the number of unique sequences to a new cell in this list - will be accumulative
                total_list.append(
                    len(temp_list)
                )  ## append the total sequences amount for plotting
                counter += 1

            x_axis.append(total_list)
            y_axis.append([len(x) for x in unique_list])
            sample_names.append(samples[sample])

        results_plot = pd.DataFrame(
            {"samples": sample_names, "x_axis": x_axis, "y_axis": y_axis}
        )
        return results_plot


class RarefractionCurves:
    def __init__(
        self,
        sequencing_report,
        samples,
        region_of_interest,
        ax=None,
        font_settings={},
        legend_settings={},
        fraction_column="cloneFraction",
    ):
        self.ax = ax
        self.plot_data = self.prepare_data(
            sequencing_report, samples, region_of_interest, fraction_column
        )
        self.plot()
        self.customize_axis(font_settings)
        self.legend(legend_settings, samples)
        self.title(font_settings)

    @staticmethod
    def prepare_data(sequencing_report, samples, region_of_interest, fraction_column):
        return PrepareData().tidy(
            sequencing_report,
            samples,
            region_of_interest,
            fraction_column=fraction_column,
        )

    def plot(self):
        if self.ax != None:
            for i, row in self.plot_data.iterrows():
                self.ax.plot(
                    row["x_axis"],
                    row["y_axis"],
                    label=row["samples"],
                    alpha=1,
                    fillstyle="full",
                    linewidth=0.8,
                )

    @staticmethod
    def customize_axis(font_settings):
        if font_settings != {}:
            plt.xlabel("Total sampled sequences", **font_settings)
            plt.ylabel("Total Unique Sequences", **font_settings)

    def legend(self, legend_settings, samples):
        if legend_settings != {}:
            plt.legend(samples, title="Sample Names", **legend_settings)

    @staticmethod
    def title(font_settings):
        if font_settings != {}:
            original_fontsize = font_settings["fontsize"]
            font_settings["fontsize"] = 22

            title = "\n".join(wrap("Sequencing depth of the given samples", 40))
            plt.title(title, pad=12, **font_settings)
            font_settings["fontsize"] = original_fontsize
